surface superSprite (
	string texturename = "";
	string texFilter = "box";
	float texBlur = 0;
	float texBlurWidth = 1;
	float premult = 1;
	float colorAccum = 0;
	float colorAmount = 0;
	string colorMode = "Mix"; // 0:Mix; 1:Add; 2:Multiply; 3:Over; 4:Screen;
	uniform float texPadding = 0;
	uniform float spriteNumPP = 1;
	float useLight = 0;
	string backLight = "normal";
	float realDisplace = 0;
	string dispTexturename = "";
	string dispFilter = "box";
	float dispBlur = 0;
	float dispBlurWidth = 1;
	float dispAmount = 0;
	uniform float dispPadding = 0;
	uniform float dispSpriteNumPP = 1;
	string dispSpriteChannel = "alpha";
	output varying color Ai = 0;
)
{
	/* Declare the sprite texture alpha amd color variable */
	float Oimg = 1;
	color Cimg = 1;

	normal Nn = normalize(N);
	float displace = 0;
	normal Nf;


	/* Use Displacement */
	if (dispAmount != 0 && dispTexturename != "")
	{
		string dispname = "";
		float displace = 0;

		if (dispPadding == 0) {
			dispname = format("%s.%d.tex", dispTexturename, dispSpriteNumPP);
		} else {
			dispname = format("%s.%04d.tex", dispTexturename, dispSpriteNumPP);
		}

		/* Get Displacement by channel */
		if (dispSpriteChannel == "alpha") {
			displace = texture (dispname[3], s, t, "blur", dispBlur, "width", dispBlurWidth, "filter", dispFilter);
		} else {
			color tex = texture (dispname, s, t, "blur", dispBlur, "width", dispBlurWidth, "filter", dispFilter);
			if (dispSpriteChannel == "red") {
				displace = comp(tex, 0);
			} else if (dispSpriteChannel == "green") {
				displace = comp(tex, 1);
			} else if (dispSpriteChannel == "blue") {
				displace = comp(tex, 2);
			} else if (dispSpriteChannel == "hue") {
				displace = comp((color "hsv" tex), 0);
			} else if (dispSpriteChannel == "saturation") {
				displace = comp((color "hsv" tex), 1);
			} else if (dispSpriteChannel == "value") {
				displace = comp((color "hsv" tex), 2);
			}
		}

		/* Add displacement */
		if (realDisplace != 0)
		{
			P = P - (displace * Nn * dispAmount);
			N = calculatenormal(P);
			N = faceforward(N, I);
		} else {
			N = calculatenormal(P - displace * dispAmount * Nn);
			N = faceforward(N, I);
		}
		Nn = normalize(N);
	}

	/* If a texture name is defined, then use a texutre */
	if (texturename != "") {
		/* Create a string with the texture name based on texPadding and the spriteNumPP */
		string texname = "";
		if (texPadding == 0) {
			texname = format("%s.%d.tex", texturename, spriteNumPP);
		} else {
			texname = format("%s.%04d.tex", texturename, spriteNumPP);
		}

		/* Read the texture alpha */
		Oimg = texture (texname[3], s, t, "blur", texBlur, "width", texBlurWidth, "filter", texFilter);
		/* Read the texture color */
		if (colorAmount != 0)
			Cimg = texture (texname, s, t, "blur", texBlur, "width", texBlurWidth, "filter", texFilter);
	}


	/* Initialize the color and opacity */
	Oi = Os * Oimg;
	Ci = Cs;


	/* Use the sprite color */
	if (colorAmount != 0)
	{
		color Csprite = Ci;

  		if (colorMode == "normal") {
			Csprite = Cimg;
		} else if (colorMode == "add") {
			Csprite = Ci + Cimg;
		} else if (colorMode == "subtract") {
			Csprite = Ci - Cimg;
		} else if (colorMode == "mult") {
			Csprite = Ci * Cimg;
		} else if (colorMode == "overlay") {
			float _Cimg, _Ci;
			float ii;
			for (ii = 0; ii < 3; ii += 1) {
				_Cimg = comp(Cimg, ii);
				_Ci = comp(Ci, ii);
				setcomp( Csprite, ii, ( (_Cimg < 0.5) ? (2 * _Ci * _Cimg) : (1 - (2 * (1 - _Ci) * (1 - _Cimg))) ) );
			}
		} else if (colorMode == "screen") {
			Csprite = 1 - ((1 - Ci) * (1 - Cimg));
		} else if (colorMode == "hardlight") {
			float _Cimg, _Ci;
			float ii;
			for (ii = 0; ii < 3; ii += 1) {
				_Cimg = comp(Cimg, ii);
				_Ci = comp(Ci, ii);
				setcomp( Csprite, ii, ( (_Cimg < 0.5) ? (2 * _Ci * _Cimg) : (1 - ((2 * (1 - _Ci) * (1 - _Ci)))) ) );
			}
		} else if (colorMode == "colburn") {
			Csprite = 1 - ((1 - Ci) / Cimg);
		} else if (colorMode == "coldodge") {
			Csprite = Ci / (1 - Cimg);
		} else if (colorMode == "darken") {
			Csprite = min (Ci, Cimg);
		} else if (colorMode == "lighten") {
			Csprite = max (Ci, Cimg);
		} else if (colorMode == "hue") {
			color _C = color "hsv" Ci;
			setcomp(_C, 0, comp((color "hsv" Cimg), 0));
			Csprite = _C;
		} else if (colorMode == "sat") {
			color _C = color "hsv" Ci;
			setcomp(_C, 1, comp((color "hsv" Cimg), 1));
			Csprite = _C;
		} else if (colorMode == "color") {
			color _C = color "hsv" Ci;
			setcomp(_C, 0, comp((color "hsv" Cimg), 0));
			setcomp(_C, 1, comp((color "hsv" Cimg), 1));
			Csprite = _C;
		} else if (colorMode == "luminosity") {
			color _C = color "hsl" Ci;
			setcomp(_C, 2, comp((color "hsl" Cimg), 2));
			Csprite = _C;
		} else {
			Csprite = Ci;
		}

		/* Mix the calculated sprite color with the Ci color */
		Ci = mix(Ci, Csprite, colorAmount);
	}


	/* Calculate premultiplication */
	if (premult == 0) {
		Ci *= Oimg;
	} else {
		Ci *= Oi;
	}


	/* Add light influence */
	if (useLight != 0)
	{
		color Clights = 0;
		illuminance(P, Nn, PI)
		{
			if (dispAmount != 0 && dispTexturename != "")
			{
				Nf = Nn;
			} else {
				Nf = normalize(faceforward(N, I));
			}
			float C = (Nf . normalize(L));
			color Cc = C;
			if (C < 0)
			{
				if (backLight == "ignore") {
					Cc = 0;
				} else if (backLight == "silhouette") {
//					Cc = (2 - (1 - C)) * (1 - Oi);
					Cc = clamp(-C, 0, 1) * (1 - Oi);
				}
			}
			Clights += Cl * Cc;
		}

		Ci *= Clights;
	}


	/* output alpha into seperate channel */
	Ai = Oi;

	/* if color accumulation is on, set the alpha to 0 */
	if (colorAccum != 0)
		Oi = 0;
}
