/*****************************************************************************************
    superSprite.sl by Ofer Zelichover (c) 2/2005
    
    This is a sprite shader for renderMan complient renderers.
	Parameters:
	Texture parameters:
	 	string texturename 	- the texture base file name. to this base file name will be added ".$.ext" where $ is the spriteNumPP value and ext is the textureext value.
		string textureext 	- the texture file name extension.
		string texFilter 	- the texture filter
		float texBlur 		- the texture blur as used in texture("textname.tex", blur)
		float texBlurWidth 	- the texture blur width
		float texPadding 	- determines the padding used for the spriteNumPP in the texture file name.
		float spriteNumPP 	- the sprite number to use for the texture file name.

	General color parameters:
	float premult 		- premultiply the texture.
	float colorAccum 	- use color accumulation. if this is 1 the the alpha channel for the sprite is opaque, in which case you can get the real alpha by using the output variable Ai.
	float useLight 		- will cause the lights in the scene to be calculated.
	float wrappedAngle 	- the wrap angle of the wrapped diffuse.

	Textuer color parameters
		float colorAmount 	- the mix amount of the texture color. when this is used, the color in the RGB channels of the texture will be compined with Cs based on the mix mode.
		string colorMode 	- the mix mode for the texture color. this is like photoshop's blending modes.

	Displacement parameters
		float realDisplace 	- if 1 then real displacement is used, if 0 then bump is used.
		string dispTexturename 	- the displacement texture file name. follows the same basic princeples as the texture file name.
		string dispFilter 		- the disp. texture filter.
		float dispBlur 			- the disp. texture blur
		float dispBlurWidth 	- the disp. texture blur width
		float dispAmount 		- the amount to displace
		float dispPadding 		- same as texPadding only for the disp. texture file name.
		float dispSpriteNumPP 	- same as spriteNumPP only for the disp. texture file name.
		string dispSpriteChannel - what channel of the disp. texture to use for the displacement. (alpha,red,green,blue,hue,sturation,value)
	
		
		output varying color Ai - holds the real alpha value. can be used in case colorAccum is used, since then the alpha channel is opaque.

		
    This shader is free to use. It comes with no warranty what-so-ever.
    You can distribute it freely, but please keep this comment block intact.
    Last modified: 2005-02-09
******************************************************************************************/

surface superSprite (
	string texturename = "";
	string textureext = "tex";
	string texFilter = "box";
	float texBlur = 0;
	float texBlurWidth = 1;
	float texNoiseAmp = 0;
	float texNoiseFreq = 4;
	float premult = 1;
	float colorAccum = 0;
	float colorAmount = 0;
	string colorMode = "Mix"; 
	uniform float texPadding = 0;
	uniform float spriteNumPP = 1;
	float useLight = 0;
	float wrappedAngle = 150;
	string backLight = "normal";
	float silhouetteMin = 0.3;
	float silhouetteMax = 0.6;
	float realDisplace = 0;
	string dispTexturename = "";
	string dispFilter = "box";
	float dispBlur = 0;
	float dispBlurWidth = 1;
	float dispAmount = 0;
	uniform float dispPadding = 0;
	uniform float dispSpriteNumPP = 1;
	string dispSpriteChannel = "alpha";
	output varying color Ai = 0;
)
{
	/* Declare the color mode functions */
	/* Over color addition */
	color overAdd (color Cunder; color Cover;)
	{
		return (Cunder + Cover);
	}
	
	/* Over color subtract */
	color overSubtract (color Cunder; color Cover;)
	{
		return (Cunder - Cover);
	}
	
	/* Over multiplication */
	color overMult (color Cunder; color Cover;)
	{
		return (Cunder * Cover);
	}
	
	/* Over overlay */
	color overOverlay (color Cunder; color Cover;)
	{
		color col = 0;
		float _under, _over;
		float ii;
		for (ii = 0; ii < 3; ii += 1) {
			_under = comp(Cunder, ii);
			_over = comp(Cover, ii);
			setcomp( col, ii, ( (_under < 0.5) ? (2 * _over * _under) : (1 - (2 * (1 - _over) * (1 - _under))) ) );
		}
		return col;
	}
	
	/* Over Screen */
	color overScreen (color Cunder; color Cover;)
	{
		return (1 - ((1 - Cover) * (1 - Cunder)));
	}

	/* Over hard light */
	color overHardlight (color Cunder; color Cover;)
	{
		color col = 0;
		float _under, _over;
		float ii;
		for (ii = 0; ii < 3; ii += 1) {
			_under = comp(Cunder, ii);
			_over = comp(Cover, ii);
			setcomp( col, ii, ( (_under < 0.5) ? (2 * _over * _under) : (1 - ((2 * (1 - _over) * (1 - _over)))) ) );
		}
		return col;
	}
	
	/* Over burn */
	color overBurn (color Cunder; color Cover;)
	{
		return (1 - ((1 - Cover) / Cunder));
	}
	
	/* Over Dodge */
	color overDodge (color Cunder; color Cover;)
	{
		return (Cover / (1 - Cunder));
	}
	
	/* Over darken */
	color overDarken (color Cunder; color Cover;)
	{
		color C;
		float Iunder = comp((color "hsv" Cunder), 2);
		float Iover = comp((color "hsv" Cover), 2);
		if (Iunder < Iover)  
			C = Cunder;
		else 
			C = Cover;
		return C;
//		return (min (Cover, Cunder));
	}	
	
	/* Over lighten */
	color overLighten (color Cunder; color Cover;)
	{
		color C;
		float Iunder = comp((color "hsv" Cunder), 2);
		float Iover = comp((color "hsv" Cover), 2);
		if (Iunder > Iover) 
			C = Cunder;
		else 
			C = Cover;
		return C;
//		return (max (Cover, Cunder));
	}	

	/* Over hue */
	color overHue (color Cunder; color Cover;)
	{
		color col = color "hsv" Cover;
		setcomp(col, 0, comp((color "hsv" Cunder), 0));
		return col;
	}

	/* Over sat */
	color overSat (color Cunder; color Cover;)
	{
		color col = color "hsv" Cover;
		setcomp(col, 0, comp((color "hsv" Cunder), 1));
		return col;
	}

	/* Over color mode */
	color overColor (color Cunder; color Cover;)
	{
		color col = color "hsv" Cover;
		setcomp(col, 0, comp((color "hsv" Cunder), 0));
		setcomp(col, 1, comp((color "hsv" Cunder), 1));
		return col;
	}
	
	/* Over luminosity mode */
	color overLum (color Cunder; color Cover;)
	{
		color col = color "hsl" Cover;
		setcomp(col, 2, comp((color "hsl" Cunder), 2));
		return col;
	}
			
	
	
	/* Declare the sprite texture alpha amd color variable */
	float Oimg = 1;
	color Cimg = 1;

	normal Nn = normalize(N);
	float displace = 0;
	normal Nf;
	
	float ss = s;
	float tt = t;

	/* Apply Noise to s and t */
	if (texNoiseAmp > 0) 
	{
		point n = noise (texNoiseFreq * P) * point(2,2,0);
		ss = s + ((1.0 - comp(n,0))  * texNoiseAmp);
		tt = t + ((1.0 - comp(n,1))  * texNoiseAmp);
	}
	
	/* Use Displacement */
	if (dispAmount != 0 && dispTexturename != "")
	{
		string dispname = "";
		float displace = 0;

		if (dispPadding == 0) {
			dispname = format("%s.%d.%s", dispTexturename, dispSpriteNumPP, textureext);
		} else {
			dispname = format("%s.%04d.%s", dispTexturename, dispSpriteNumPP, textureext);
		}

		/* Get Displacement by channel */
		if (dispSpriteChannel == "alpha") {
			displace = texture (dispname[3], ss, tt, "blur", dispBlur, "width", dispBlurWidth, "filter", dispFilter);
		} else {
			color tex = texture (dispname, ss, tt, "blur", dispBlur, "width", dispBlurWidth, "filter", dispFilter);
			if (dispSpriteChannel == "red") {
				displace = comp(tex, 0);
			} else if (dispSpriteChannel == "green") {
				displace = comp(tex, 1);
			} else if (dispSpriteChannel == "blue") {
				displace = comp(tex, 2);
			} else if (dispSpriteChannel == "hue") {
				displace = comp((color "hsv" tex), 0);
			} else if (dispSpriteChannel == "saturation") {
				displace = comp((color "hsv" tex), 1);
			} else if (dispSpriteChannel == "value") {
				displace = comp((color "hsv" tex), 2);
			}
		}

		/* Add displacement */
		if (realDisplace != 0)
		{
			P += (displace * Nn * dispAmount);
			N = calculatenormal(P);
			N = faceforward(N, I);
		} else {
			N = calculatenormal(P + displace * dispAmount * Nn);
			N = faceforward(N, I);
		}
		Nn = normalize(N);
	}

	/* If a texture name is defined, then use a texutre */
	if (texturename != "") 
	{
		/* Create a string with the texture name based on texPadding and the spriteNumPP */
		string texname = "";
		if (texPadding == 0) {
			texname = format("%s.%d.%s", texturename, spriteNumPP, textureext);
		} else if (texPadding == 2) {
			texname = format("%s.%02d.%s", texturename, spriteNumPP, textureext);
		} else if (texPadding == 3) {
			texname = format("%s.%03d.%s", texturename, spriteNumPP, textureext);
		} else if (texPadding == 5) {
			texname = format("%s.%05d.%s", texturename, spriteNumPP, textureext);
		} else if (texPadding == 6) {
			texname = format("%s.%06d.%s", texturename, spriteNumPP, textureext);
		} else {
			texname = format("%s.%04d.%s", texturename, spriteNumPP, textureext);
		}

		/* Read the texture alpha */
		Oimg = texture (texname[3], ss, tt, "blur", texBlur, "width", texBlurWidth, "filter", texFilter);
		/* Read the texture color */
		if (colorAmount != 0)
			Cimg = texture (texname, ss, tt, "blur", texBlur, "width", texBlurWidth, "filter", texFilter);
	}


	/* Initialize the color and opacity */
	Oi = Os * Oimg;
	Ci = Cs;


	/* Use the sprite color */
	if (colorAmount != 0)
	{
		color Csprite = Ci;

  		if (colorMode == "normal") {
			Csprite = Cimg;
		} else if (colorMode == "add") {
			Csprite = overAdd (Ci, Cimg);
		} else if (colorMode == "subtract") {
			Csprite = overSubtract (Ci, Cimg);
		} else if (colorMode == "mult") {
			Csprite = overMult (Ci, Cimg);
		} else if (colorMode == "overlay") {
			Csprite = overOverlay (Ci, Cimg);
		} else if (colorMode == "screen") {
			Csprite = overScreen (Ci, Cimg);
		} else if (colorMode == "hardlight") {
			Csprite = overHardlight (Ci, Cimg);
		} else if (colorMode == "colburn") {
			Csprite = overBurn (Ci, Cimg);
		} else if (colorMode == "coldodge") {
			Csprite = overDodge (Ci, Cimg);
		} else if (colorMode == "darken") {
			Csprite = overDarken (Ci, Cimg);
		} else if (colorMode == "lighten") {
			Csprite = overLighten (Ci, Cimg);
		} else if (colorMode == "hue") {
			Csprite = overHue (Ci, Cimg);
		} else if (colorMode == "sat") {
			Csprite = overSat (Ci, Cimg);
		} else if (colorMode == "color") {
			Csprite = overColor (Ci, Cimg);
		} else if (colorMode == "luminosity") {
			Csprite = overLum (Ci, Cimg);
		} else {
			Csprite = Ci;
		}

		/* Mix the calculated sprite color with the Ci color */
		Ci = mix(Ci, Csprite, colorAmount);
	}


	/* Calculate premultiplication */
	if (premult == 0) {
		Ci *= Oimg;
	} else {
		Ci *= Oi;
	}


	/* Add light influence */
	if (useLight != 0)
	{
		color Clights = 0;
		illuminance(P, Nn, PI)
		{
			if (dispAmount != 0 && dispTexturename != "")
			{
				Nf = Nn;
			} else {
				Nf = normalize(faceforward(N, I));
			}
			float C = 1 - acos(Nf . normalize(L)) / wrappedAngle;

			if ((Nf . normalize(L)) < 0)
			{
				float Ov = comp((color "hsv" Oi), 2);
				if (backLight == "ignore") {
					C = 0;
				} else if (backLight == "silhouette") {
					C = clamp(-C, 0, 1) * smoothstep(silhouetteMin, silhouetteMax, (1.-Ov));
				}
			}
			Clights += Cl * C;
		}

		Ci *= Clights;
	}


	/* output alpha into seperate channel */
	Ai = Oi;

	/* if color accumulation is on, set the alpha to 0 */
	if (colorAccum != 0)
		Oi = 0;
}
