--*********************************************************************************
--* This file is part of the RIBit Renderman(tm) exporter for max
--* By Ofer Zelichover
--* 2004/3/1 (c)
--* based in part on kcman by Rudy Cortes

--*********************************************************************************


struct s_RIBit_Camera (
	cam,
	shadersToComplie = #(),
	shadersDir = "",


	settings = s_RIBit_Settings(),
	indent = "",
	fileHandle,



	fn exportRIBBox =
	(
		global RIBit_Functions
		if isValidNode cam then (
			local modIndex = RIBit_Functions.findModifier cam mo_RIBit_RIBBox multiple:true
			for i in modIndex do (
				format "%##---- RIB Box ----##\n" indent to:fileHandle
				local lines = filterString (RIBit_Functions.parseRIBBox cam.modifiers[i].RIBCode params:#(cam)) "\n\r"
				for l in lines do
					format "%%\n" indent l to:fileHandle
			)
		)
	),


	fn exportRenderingAttributes =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local modIndex = RIBit_Functions.findModifier cam mo_RIBit_Shaders
		if modIndex != undefined then (
			local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle

		-- color
			fn motionFunc params: =
			(
				params[1].params = #()
				params[1].addParameter (getProperty params[2] params[3]) undefined
				params[1].write()
			)

			local ribColor = s_RIBit_Option "Color" indent:indent fileHandle:fileHandle
			ribMotion.func = motionFunc
			ribMotion.funcParams = #(ribColor, cam.modifiers[modIndex], "Color")

			ribMotion.dependsOn = #(cam.modifiers[modIndex].color.isAnimated)
			ribMotion.bypass = settings.useConservativeMotion
			ribMotion.write()
		-- Opacity
			ribColor.name = "Opacity"
			ribMotion.funcParams = #(ribColor, cam.modifiers[modIndex], "Opacity")

			ribMotion.dependsOn = #(cam.modifiers[modIndex].opacity.isAnimated)
			ribMotion.write()

		-- surface
			local ribShader = s_RIBit_Option "Surface" indent:indent fileHandle:fileHandle
			fn motionFunc params: =
			(
				global RIBit_Functions
				params[1].params = #()
				params[1].addParameter params[2] undefined
				params[1].paramsFromText (RIBit_Functions.parseRIBBox params[3] params:#(params[4]))
				params[1].write()
			)
			ribMotion.func = motionFunc
			if cam.modifiers[modIndex].Surface != "" then (
				ribMotion.funcParams = #(ribShader, cam.modifiers[modIndex].Surface, cam.modifiers[modIndex].SurfaceParams, cam)
				ribMotion.dependsOn = #(cam.modifiers[modIndex].SurfaceParams != "")
				ribMotion.write()
				local shaderSource = shadersDir + cam.modifiers[modIndex].Surface
				if doesFileExist (shaderSource + ".sl") then
					append shadersToComplie #((shaderSource + ".sl"), shaderSource)
			)

			if cam.modifiers[modIndex].Displacement != "" then (
				ribShader.name = "Displacement"

			ribMotion.funcParams = #(ribShader, cam.modifiers[modIndex].Displacement, cam.modifiers[modIndex].DisplacementParams, cam)
				ribMotion.dependsOn = #(cam.modifiers[modIndex].DisplacementParams != "")
				ribMotion.write()
				local shaderSource = shadersDir + cam.modifiers[modIndex].Displacement
				if doesFileExist (shaderSource + ".sl") then
					append shadersToComplie #((shaderSource + ".sl"), shaderSource)
			)
			if cam.modifiers[modIndex].Atmosphere != "" then(
				ribShader.name = "Atmosphere"
				ribMotion.funcParams = #(ribShader, cam.modifiers[modIndex].Atmosphere, cam.modifiers[modIndex].AtmosphereParams, cam)
				ribMotion.dependsOn = #(cam.modifiers[modIndex].AtmosphereParams!= "")
				ribMotion.write()
				local shaderSource = shadersDir + cam.modifiers[modIndex].Atmosphere
				if doesFileExist (shaderSource + ".sl") then
					append shadersToComplie #((shaderSource + ".sl"), shaderSource)
			)
			format "%ShadingInterpolation \"%\"\n" indent cam.modifiers[modIndex].ShadingInterpolation to:fileHandle
			format "%Matte %\n" indent (if cam.modifiers[modIndex].Matte then 1 else 0) to:fileHandle
		)
	),


	fn expotLightAsCamera =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		local ribProjection = s_RIBit_Option "Projection" indent:indent fileHandle:fileHandle
		local nearClip = 0.1
		local farClip = 100000.0
		if isKindOf cam TargetDirectionallight or isKindOf cam Directionallight then (
			ribProjection.addParameter "orthographic" undefined
			ribProjection.write()
			local xmin = -cam.falloff
			local xmax = cam.falloff
			local ymin = -(cam.falloff / cam.aspect)
			local ymax = (cam.falloff / cam.aspect)
			Format "%ScreenWindow % % % %\n" indent xmin xmax ymin ymax to:fileHandle
		) else (
			ribProjection.addParameter "perspective" undefined

			fn motionFunc params: =
			(
				local cameraVFOV = if (isKindOf params[2] freeSpot or isKindOf params[2] targetSpot) then params[2].falloff else 95
				params[1].changeValue "fov" cameraVFOV
				params[1].write()
			)
			ribMotion.func = motionFunc
			ribMotion.funcParams = #(ribProjection, cam)
			ribMotion.dependsOn = #(cam.fov.isAnimated)
			ribMotion.bypass = settings.useConservativeMotion
			ribMotion.write()
		)
		Format "%Clipping % %\n" indent nearClip farClip to:fileHandle

		fn motionFunc params: =
		(
			params[1].transform = params[2].transform
			params[1].inversePOV()
			params[1].write()
		)
		local ribTransform = s_RIBit_Transform indent:indent fileHandle:fileHandle
		ribMotion.func = motionFunc
		ribMotion.funcParams = #(ribTransform, cam)
		ribMotion.dependsOn = #(cam.transform.isAnimated)
		ribMotion.bypass = false
		ribMotion.write()
	),



	fn exportCamera =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		local renderAspect = (settings.rendWidth as float)/settings.rendHeight
		local ribTransform = s_RIBit_Transform (cam.transform) indent:indent fileHandle:fileHandle

		local ribProjection = s_RIBit_Option "Projection" indent:indent fileHandle:fileHandle

		if cam.orthoProjection then (
			fn motionFunc params: =
			(
				params[1].transform = params[2].transform
				params[1].inversePOV()
				params[1].write()
			)
			ribProjection.addParameter "orthographic" undefined
			ribProjection.write()
		) else (
			ribProjection.addParameter "perspective" undefined

			fn motionFunc params: =
			(
				local VFOV = if params[3] > 1 then (2.*atan(tan(params[2].fov/2.)/params[3])) else params[2].fov
				params[1].changeValue "fov" VFOV
				params[1].write()
			)
			ribMotion.func = motionFunc
			ribMotion.funcParams = #(ribProjection, cam, renderAspect)
			ribMotion.dependsOn = #(cam.fov.isAnimated)
			ribMotion.bypass = settings.useConservativeMotion
			ribMotion.write()

			fn motionFunc params: =
			(
				params[1].transform = params[2].transform
				params[1].inversePOV()
				params[1].write()
			)
			ribMotion.func = motionFunc
			ribMotion.funcParams = #(ribTransform, cam)
			ribMotion.dependsOn = #(cam.transform.isAnimated)
			ribMotion.bypass = false
		)

		if settings.useDOF then (
			local cameraVFOV = if renderAspect > 1 then (2.*atan(tan(cam.fov/2.)/renderAspect)) else cam.fov
			local fLength = renderAspect / tan (cameraVFOV / 2.)
			local fDist = cam.baseObject.targetDistance
			Format "%DepthOfField % % %\n" indent settings.DOFAperture fLength fDist to:fileHandle
		)
		if cam.clipManually then
			Format "%Clipping % %\n" indent cam.nearClip cam.farClip to:fileHandle

		ribMotion.write()
	),


	fn exportViewport =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local viewtype = viewport.getType()
		local ribProjection = s_RIBit_Option "Projection" indent:indent fileHandle:fileHandle
		if (viewtype == #view_camera) then (
			cam = viewport.getCamera()
			exportCamera()
		) else if (viewtype == #view_persp_user) then (
			--- code from doberman
			local renderAspect = (settings.rendWidth as float)/settings.rendHeight
			local cameraVFOV = if renderAspect > 1 then (2.*atan(tan(getViewFOV()/2.)/renderAspect)) else getViewFOV()
			local nearClip = 0.1
			local farClip = 100000.0
			if cam != undefined then (
				if cam.clipManually == true then
				(
					local nearClip = cam.nearclip
					local farClip = cam.farclip
				)
			)
			ribProjection.addParameter "perspective" undefined
			ribProjection.addParameter "fov" cameraVFOV
			ribProjection.write()
			Format "%Clipping % %\n" indent nearClip farClip to:fileHandle

			local ribTransform = s_RIBit_Transform (inverse (getViewTM())) indent:indent fileHandle:fileHandle
			ribTransform.inversePOV()
			ribTransform.write()
		) else if (viewtype == #view_iso_user) then (
format "RIBit::Camera::exportViewport::viewtype=%\n" viewtype
			local nearClip = 0.1
			local farClip = 100000.0
			local renderAspect = (settings.rendWidth as float)/settings.rendHeight
			local cameraVFOV = if renderAspect > 1 then (2.*atan(tan(getViewFOV()/2.)/renderAspect)) else getViewFOV()

			ribProjection.addParameter "orthographic" undefined
			ribProjection.write()

			Format "%Clipping % %\n" indent nearClip farClip to:fileHandle
			local width = (getScreenScaleFactor (getViewTM())[4]) * tan (cameraVFOV / 2.) --/ renderAspect
			local xmin = -(2. * width) -- -(winSize) * 2. --/ renderAspect
			local xmax = 2. * width --(winSize) * 2. --/ renderAspect
			local ymin = -(width / renderAspect) -- -(winSize) / renderAspect
			local ymax = width / renderAspect --(winSize) / renderAspect
			Format "%ScreenWindow % % % %\n" indent xmin xmax ymin ymax to:fileHandle

			local ribTransform = s_RIBit_Transform ( (getViewTM())) indent:indent fileHandle:fileHandle
--			ribTransform.transform = translate ribTransform.transform [0,0,(getScreenScaleFactor (getViewTM())[4])]
--		ribTransform.transform = inverse ribTransform.transform
--		ribTransform.transform = rotateY ribTransform.transform 180
--		ribTransform.transform *= scaleMatrix [-1,1,1]

--			ribTransform.inversePOV()
--			ribTransform.transform = rotateY ribTransform.transform 180
			ribMotion.write()
		) else (
--				format "%Projection \"orthographic\"\n" indent to:fileHandle
--				local r = mapScreenToWorldRay (getViewSize() / 2.)
--				local ribTransform = s_RIBit_Transform (matrixFromNormal r.dir) indent:indent fileHandle:fileHandle
--				ribTransform.inversePOV()
--		transform = inverse transform
--		ribTransform.transform = translate ribTransform.transform r.pos
--		ribTransform.transform = inverse ribTransform.transform
--		ribTransform.inversePOV()
--		scale [-1,1,1]
--				ribTransform.write()

		)
	),



	fn write =
	(
		if fileHandle == undefined then return()
		case (superClassOf cam) of (
			-- if cam is a light, we're exporting a shadowmap RIB.
			light: expotLightAsCamera()
			-- if cam is a camera object, use it.
			camera: (exportRenderingAttributes (); exportCamera())
			-- otherwise, export the active viewport.
			default: exportViewport()
		)
		exportRIBBox()
	)

) -- end of s_RIBit_Camera struct


