--*********************************************************************************
--* This file is part of the RIBit Renderman(tm) exporter for max
--* By Ofer Zelichover
--* 2004/3/1 (c)
--* based in part on kcman by Rudy Cortes

--*********************************************************************************

struct s_RIBit_Geometry (
	object,
	shadersToComplie = #(),
	shadersDir = "",
	force2sided = false,
	camObject,
	settings = s_RIBit_Settings(),
	objInstances = #(),
	indent = "",
	fileHandle,
	canceled = false,


	fn exportPointsPolygons params: =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local type = params[1]
		local object = params[2]
		local indent = params[3]
		local fileHandle = params[4]
		canceled = params[5]
		if canceled then return canceled
		canceled = getProgressCancel()
		max create mode
		with redraw off (
			------------------
			-- check whether it is a poly
			-----------------
			local obj = snapshot object
			local facecount = getNumFaces obj
			local numverts = getNumVerts obj
			
			format "%PointsPolygons \n" indent to:fileHandle

			--First output the number of vertices per face
			format "%\t[" indent to:fileHandle
			for i = 1 to facecount where (not canceled) do(
				canceled = getProgressCancel()
				if (mod i 25 == 0 )then
					format"\n%\t" indent to:fileHandle
				format " 3 " to:fileHandle
			)
			format "]\n" to:fileHandle
			
			-- now output the vertex indexes
			local vertindex = 0;
			format "%\t[" indent to:fileHandle
			for i = 1 to facecount where (not canceled) do (
				canceled = getProgressCancel()
				if (mod i 9 == 0 ) then
					format"\n%\t" indent to:fileHandle
				for j = 1 to 3 do (
					format " %" vertindex to:fileHandle
					vertindex += 1
				)
			)
			format "]\n" to:fileHandle
			
			-- now output the position of each vertex
			format "%\t\"P\" [" indent to:fileHandle
			for i = 1 to facecount where (not canceled) do(
				canceled = getProgressCancel()
				if (mod i 9 == 0 )then
					format"\n%\t" indent to:fileHandle
				local faceVerts = getFace obj i
				for f = 1 to 3 do (
					in coordsys local vertpos = getVert obj faceVerts[f]
					format " % % % " vertpos[1] vertpos[2] vertpos[3] to:fileHandle
				)
			)
			format "]\n" to:fileHandle
	
		-- now output the texture coordinates. first "s"
			local modIndex = s_RIBit_Functions.findModifier object mo_RIBit_Geometry
			if modIndex == undefined or (modIndex != undefined and object.modifiers[modIndex].ExportTexCoords) then (
				local mapChannel = if modIndex == undefined then 1 else object.modifiers[modIndex].TexCoordsChannel
				if meshOp.getMapSupport obj mapChannel then (
					-- first "s"
					format "%\t\"vertex float s\" [" indent to:fileHandle
					for i = 1 to facecount where (not canceled) do(
						canceled = getProgressCancel()
						if (mod i 25 == 0 )then
							format"\n%\t" indent to:fileHandle
							
						local faceVerts = getFace obj i
						local tface = meshOp.getMapFace obj mapChannel i         -- rjw 
						for f = 1 to 3 do (
					-- rjw		local mapCoord = in coordsys local meshOp.getMapVert obj mapChannel faceVerts[f]
							local mapCoord = in coordsys local meshOp.getMapVert obj mapChannel tface[f]  -- rjw
							format " % " mapCoord[1] to:fileHandle
						)
					)
					format "]\n" to:fileHandle
			
					-- then "t"
					format "%\t\"vertex float t\" [" indent to:fileHandle
					for i = 1 to facecount where (not canceled) do(
						canceled = getProgressCancel()
						if (mod i 25 == 0 )then
							format"\n%\t" indent to:fileHandle
							
						local faceVerts = getFace obj i
						-- next line, rjw
						local tface = meshOp.getMapFace obj mapChannel i   
						for f = 1 to 3 do (
					-- rjw		local mapCoord = in coordsys local meshOp.getMapVert obj mapChannel faceVerts[f]
							local mapCoord = in coordsys local meshOp.getMapVert obj mapChannel tface[f] -- rjw
							format " % " mapCoord[2] to:fileHandle
						)
					)
					format "]\n" to:fileHandle
				)
			)
			
			--output vertex channels
			-- -2 is the alpha channel
			-- -1 is the illumination channel
			-- 0 is the vertex color channel
			-- 1+ is the mapping channel
			local modIndex = s_RIBit_Functions.findModifier object mo_RIBit_VertexAttribute multiple:true
			for i in modIndex where (not canceled) do (
				canceled = getProgressCancel()
				local chan = object.modifiers[i].Channel
				local chanSupported = false
				case chan of (
					1: chan = 0
					2: chan = -2
					3: chan = -1
					4: chan = object.modifiers[i].mappingChannel
				)
				chanSupported = meshOp.getMapSupport obj chan
				if chanSupported then (
					local attrType = case object.modifiers[i].AttributeType of (
							1: "float"
							2: "color"
					)
					format "%\t\"vertex % %\" [" indent attrType object.modifiers[i].AttributeName to:fileHandle
					for i = 1 to facecount where (not canceled) do(
						canceled = getProgressCancel()
						if (mod v 9 == 0)then
							format"\n%\t" indent to:fileHandle
						
						local faceVerts = getFace obj i
						for f = 1 to 3 do (
							local col = meshOp.getMapVert obj chan faceVerts[f]
							case object.modifiers[i].AttributeType of (
								1: (col *= 255.; format " % " (s_RIBit_Functions.colorToFloat col[1] col[2] col[3] mode:#average) to:fileHandle)
								2: format " % % % " col[1] col[2] col[3] to:fileHandle
							)
						)
					)
					format "]\n" to:fileHandle
				)
			)
			
			-- now output the normal of each vertex
			local modIndex = s_RIBit_Functions.findModifier object mo_RIBit_Geometry
			if modIndex == undefined or object.modifiers[modIndex].ExportNormals then (
				format "%\t\"N\" [" indent to:fileHandle
				-- convertToMesh obj
				local numverts = getNumVerts obj
				for i = 1 to facecount where (not canceled) do(
					canceled = getProgressCancel()
					if (mod i 9 == 0 )then
						format"\n%\t" indent to:fileHandle
						
					in coordsys local normals = meshop.getFaceRNormals obj i
					for j = 1 to 3 do					
						format " % % % " normals[j][1] normals[j][2] normals[j][3] to:fileHandle
				)
				format "]\n" to:fileHandle
			)
	
			delete obj
		) -- end with redraw off
	),


	fn exportPolygons params: =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local type = params[1]
		local object = params[2]
		local indent = params[3]
		local fileHandle = params[4]
		canceled = params[5]
		if canceled then return canceled
		canceled = getProgressCancel()
		max create mode
		with redraw off (
			------------------
			-- check whether it is a poly
			-----------------
			local obj = snapshot object
			convertToPoly obj
			local facecount = getNumFaces obj
			local numverts = getNumVerts obj
			
			-- chace used functions
			local fnGetFaceVerts = polyop.getFaceVerts --meshOp.getVertsUsingFace
			local fnGetVert = polyop.getVert  --meshOp.getVert
			local fnGetMapVert = polyop.getMapVert  --meshOp.getMapVert
			local fnGetFacesUsingVert = polyOp.getFacesUsingVert  --meshOp.getFacesUsingVert
			local fnGetMapSupport = polyOp.getMapSupport  --meshOp.getMapSupport
			local fnGetNormal = getNormal  --polyOp.getFaceNormal


			case type of (
				1: format "%PointsPolygons \n" indent to:fileHandle
				2: format "%SubdivisionMesh \"catmull-clark\"\n" indent to:fileHandle
			)

			--if ((mode == 1) or (mode == 2)) then(
			--First output the number of vertices per face
			format "%\t[" indent to:fileHandle
			for i=1 to facecount where (not canceled) do(
				canceled = getProgressCancel()
				if (mod i 25 == 0 )then
					format"\n%\t" indent to:fileHandle
				local faceverts = fnGetFaceVerts obj i
				local numverts = faceverts.count  --faceverts.numberSet
				format " % " numverts to:fileHandle
			)
			format "]\n" to:fileHandle

			-- now output the vertex indexes
			format "%\t[" indent to:fileHandle
			for i=1 to facecount where (not canceled) do(
				canceled = getProgressCancel()
				if (mod i 9 == 0 ) then
					format"\n%\t" indent to:fileHandle
				local faceverts = fnGetFaceVerts obj i
				for j in faceverts do(
					format " % " (j - 1) to:fileHandle
				)
			)
			format "]\n" to:fileHandle

			-- now output the position of each vertex
			format "%\t\"P\" [" indent to:fileHandle
			for i = 1 to numverts do (
				canceled = getProgressCancel()
				if (mod i 9 == 0 )then
					format"\n%\t" indent to:fileHandle
				vertpos = in coordsys local fnGetVert obj i
				format " % % % " vertpos[1] vertpos[2] vertpos[3] to:fileHandle
			)
			format "]\n" to:fileHandle

			-- now output the texture coordinates. first "s"
			local modIndex = RIBit_Functions.findModifier object mo_RIBit_Geometry
--			if modIndex == undefined or (modIndex != undefined and object.modifiers[modIndex].ExportTexCoords) then (
			if (modIndex != undefined and object.modifiers[modIndex].ExportTexCoords) then (
				local mapChannel = if modIndex == undefined then 1 else object.modifiers[modIndex].TexCoordsChannel
				if fnGetMapSupport obj mapChannel then (
					-- first "s"
					format "%\t\"vertex float s\" [" indent to:fileHandle
					for i = 1 to numverts where (not canceled) do (
						canceled = getProgressCancel()
						if (mod i 25 == 0 )then
							format"\n%\t" indent to:fileHandle
						local mapCoord = in coordsys local fnGetMapVert obj mapChannel i
						format " % " mapCoord[1] to:fileHandle
					)
					format "]\n" to:fileHandle

					-- then "t"
					format "%\t\"vertex float t\" [" indent to:fileHandle
					for i = 1 to numverts where (not canceled) do (
						canceled = getProgressCancel()
						if (mod i 25 == 0 )then
							format"\n%\t" indent to:fileHandle
						local mapCoord = in coordsys local fnGetMapVert obj mapChannel i
						format " % " mapCoord[2] to:fileHandle
					)
					format "]\n" to:fileHandle
				)
			)

			--output vertex channels
			-- -2 is the alpha channel
			-- -1 is the illumination channel
			-- 0 is the vertex color channel
			-- 1+ is the mapping channel
			local modIndex = RIBit_Functions.findModifier object mo_RIBit_VertexAttribute multiple:true
			for i in modIndex where (not canceled) do (
				canceled = getProgressCancel()
				local chan = object.modifiers[i].Channel
				local chanSupported = false
				case chan of (
					1: chan = 0
					2: chan = -2
					3: chan = -1
					4: chan = object.modifiers[i].mappingChannel
				)
				chanSupported = fnGetMapSupport obj chan
				if chanSupported then (
					local attrType = case object.modifiers[i].AttributeType of (
							1: "float"
							2: "color"
						)
					format "%\t\"vertex % %\" [" indent attrType object.modifiers[i].AttributeName to:fileHandle
					for v = 1 to numverts where (not canceled) do (
						canceled = getProgressCancel()
						if (mod v 9 == 0)then
							format"\n%\t" indent to:fileHandle
						local col = fnGetMapVert obj chan v
						case object.modifiers[i].AttributeType of (
							1: (col *= 255.; format " % " (RIBit_Functions.colorToFloat col[1] col[2] col[3] mode:#average) to:fileHandle)
							2: format " % % % " col[1] col[2] col[3] to:fileHandle
						)
					)
					format "]\n" to:fileHandle
				)
			)
/*
			-- now output the normal of each vertex
			local modIndex = RIBit_Functions.findModifier object mo_RIBit_Geometry
			if modIndex == undefined or object.modifiers[modIndex].ExportNormals then (
				format "%\t\"N\" [" indent to:fileHandle
--				convertToMesh obj
--				meshop.unifyNormals obj #{1..obj.numFaces}
--				local numverts = getNumVerts obj
				for i = 1 to numverts where (not canceled) do (
					canceled = getProgressCancel()
					if (mod i 9 == 0 )then
						format"\n%\t" indent to:fileHandle
					normpos = [0,0,0]
					local faces = fnGetFacesUsingVert obj i
					for f in faces do
						in coordsys object normpos += fnGetNormal obj f
					normpos /= faces.numberSet
					format " % % % " normpos[1] normpos[2] normpos[3] to:fileHandle
				)
				format "]\n" to:fileHandle
			)
*/
			-- now output the normal of each vertex
			local modIndex = RIBit_Functions.findModifier object mo_RIBit_Geometry
--			if modIndex == undefined or object.modifiers[modIndex].ExportNormals then (
			if modIndex != undefined and object.modifiers[modIndex].ExportNormals then (
				convertToMesh obj
				format "%\t\"N\" [" indent to:fileHandle
--				local facecount = getNumFaces obj
				for i=1 to numverts where (not canceled) do(
					canceled = getProgressCancel()
					if (mod i 9 == 0 )then
						format"\n%\t" indent to:fileHandle
						
					in coordsys local normals = fnGetNormal obj i
					format " % % % " normals[1] normals[2] normals[3] to:fileHandle
--					format " % % % " normals[2][1] normals[2][2] normals[2][3] to:fileHandle
--					format " % % % " normals[3][1] normals[3][2] normals[3][3] to:fileHandle
				)
				format "]\n" to:fileHandle
			)
			
			delete obj
		) -- end with redraw off
	),


	fn exportRenderingAttributes =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		if canceled then return()
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Shaders
		if modIndex == undefined then (
			if object.parent == undefined and (not isValidNode camObject or RIBit_Functions.findModifier camObject mo_RIBit_Shaders == undefined) then (
				local objcolor = object.wirecolor/255.
				format "%Color % % %\n" indent objcolor.r objcolor.g objcolor.b to:fileHandle
				format "%Opacity % % %\n" indent objcolor.a objcolor.a objcolor.a to:fileHandle
				format "%Surface \"matte\"\n" indent to:fileHandle
			)
		) else (
			local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle

		-- color
			fn motionFunc params: =
			(
				params[1].params = #()
				params[1].addParameter (getProperty params[2] params[3]) undefined
				params[1].write()
			)

			local ribColor = s_RIBit_Option "Color" indent:indent fileHandle:fileHandle
			ribMotion.func = motionFunc
			ribMotion.funcParams = #(ribColor, object.modifiers[modIndex], "Color")

			ribMotion.dependsOn = #(object.modifiers[modIndex].color.isAnimated)
			ribMotion.bypass = settings.useConservativeMotion
			ribMotion.write()
		-- Opacity
			ribColor.name = "Opacity"
			ribMotion.funcParams = #(ribColor, object.modifiers[modIndex], "Opacity")

			ribMotion.dependsOn = #(object.modifiers[modIndex].opacity.isAnimated)
			ribMotion.write()

		-- surface
			local ribShader = s_RIBit_Option "Surface" indent:indent fileHandle:fileHandle
			fn motionFunc params: =
			(
				params[1].params = #()
				params[1].addParameter params[2] undefined
				params[1].paramsFromText (RIBit_Functions.parseRIBBox params[3] params:#(params[4]))
				params[1].write()
			)
			ribMotion.func = motionFunc
			if object.modifiers[modIndex].Surface != "" then (
				ribMotion.funcParams = #(ribShader, object.modifiers[modIndex].Surface, object.modifiers[modIndex].SurfaceParams, object)
				ribMotion.dependsOn = #(object.modifiers[modIndex].SurfaceParams != "")
				ribMotion.write()
				local shaderSource = shadersDir + object.modifiers[modIndex].Surface
				if doesFileExist (shaderSource + ".sl") then
					append shadersToComplie #((shaderSource + ".sl"), shaderSource)
			)

			if object.modifiers[modIndex].Displacement != "" then (
				ribShader.name = "Displacement"

			ribMotion.funcParams = #(ribShader, object.modifiers[modIndex].Displacement, object.modifiers[modIndex].DisplacementParams, object)
				ribMotion.dependsOn = #(object.modifiers[modIndex].DisplacementParams != "")
				ribMotion.write()
				local shaderSource = shadersDir + object.modifiers[modIndex].Displacement
				if doesFileExist (shaderSource + ".sl") then
					append shadersToComplie #((shaderSource + ".sl"), shaderSource)
			)
			if object.modifiers[modIndex].Atmosphere != "" then(
				ribShader.name = "Atmosphere"
				ribMotion.funcParams = #(ribShader, object.modifiers[modIndex].Atmosphere, object.modifiers[modIndex].AtmosphereParams, object)
				ribMotion.dependsOn = #(object.modifiers[modIndex].AtmosphereParams!= "")
				ribMotion.write()
				local shaderSource = shadersDir + object.modifiers[modIndex].Atmosphere
				if doesFileExist (shaderSource + ".sl") then
					append shadersToComplie #((shaderSource + ".sl"), shaderSource)
			)
			format "%ShadingInterpolation \"%\"\n" indent object.modifiers[modIndex].ShadingInterpolation to:fileHandle
			format "%Matte %\n" indent (if object.modifiers[modIndex].Matte then 1 else 0) to:fileHandle
		)
	),


	fn exportGeometry =
	(
		-- Export Polygons
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local mode = 1
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Geometry
		if modIndex == undefined then
			mode = 1
		else (
			if (object.modifiers[modIndex].ExportPoly and not object.modifiers[modIndex].ExportSubd) then
				mode = 1
			if (object.modifiers[modIndex].ExportPoly and object.modifiers[modIndex].ExportSubd) then
				mode = 2
		)

		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		case mode of (
			1: ribMotion.func = exportPointsPolygons
			2: ribMotion.func = exportPolygons
			default: ribMotion.func = exportPointsPolygons
		)
		ribMotion.funcParams = #(mode, object, indent, fileHandle, canceled)
		ribMotion.dependsOn = #(true)
		ribMotion.write()

	),


	fn exportParticles =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		if canceled then return()
		local ribParticle = s_RIBit_Particle object:object camObject:camObject settings:settings objInstances:objInstances indent:indent fileHandle:fileHandle
		ribParticle.write()
	),


	fn exportLightLink =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		if canceled then return()
		local allLights = for l in lights where ((not isKindOf l Targetobject) and (not canceled)) collect l
		for i = 1 to allLights.count do (
			canceled = getProgressCancel()

			local l = allLights[i]
			local ribIllum = s_RIBit_Option "Illuminate" indent:indent fileHandle:fileHandle
			ribIllum.addParameter i undefined

			if (l.includeList != undefined) and (findItem l.includeList object > 0) then (
				ribIllum.addParameter 1 undefined
				ribIllum.write()
			)

			if (l.excludeList != undefined) and (findItem l.excludeList object > 0) then (
				ribIllum.addParameter 0 undefined
				ribIllum.write()
			)
		)
	),


	fn exportRIBBox =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_RIBBox multiple:true
		for i in modIndex do (
			format "%##---- RIB Box ----##\n" indent to:fileHandle
			local lines = filterString (RIBit_Functions.parseRIBBox object.modifiers[i].RIBCode params:#(object)) "\n\r"
			for l in lines do
				format "%%\n" indent l to:fileHandle
		)
	),

	fn write =
	(
		if fileHandle == undefined then return()
		-- Rendering Attributes
		exportRenderingAttributes()
		exportLightLink()

		local ribTransform = s_RIBit_Transform concat:true indent:indent fileHandle:fileHandle
		ribTransform.transformBegin()
		indent = ribTransform.indent

		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		ribMotion.func = fn motionFunc params: = (params[1].transform = params[2].transform; params[1].write())
		ribMotion.funcParams = #(ribTransform, object)
		ribMotion.dependsOn = #(object.transform.isAnimated)
		ribMotion.write()

		exportRIBBox()

		if isKindOf object GeometryClass and not (isKindOf object targetObject) and object.renderable and not object.isHidden then (

			case (classOf object) of (
				Snow: 		exportParticles()
				Spray: 		exportParticles()
				Blizzard: 	exportParticles()
				PArray: 	exportParticles()
				PCloud: 	exportParticles()
				SuperSpray: exportParticles()
				PF_Source: 	exportParticles()
				default: 	exportGeometry()
			)
		)

		ribTransform.transformEnd()
		indent = ribTransform.indent
		-- End Transformation block
		if not canceled then
			canceled = getProgressCancel()
	)

) -- end of s_RIBit_Geometry struct













