--*********************************************************************************
--* This file is part of the RIBit Renderman(tm) exporter for max
--* By Ofer Zelichover
--* 2004/3/1 (c)
--* Last Update: 11:14 PM 10/10/2006
--* based in part on kcman by Rudy Cortes

--*********************************************************************************


struct s_RIBit_Parameter (
	name, value,

	fn write =
	(
		global RIBit_Functions
		local returnStr = "" as stringStream
		if name != undefined then
			format "%" (RIBit_Functions.writeValueByType name) to:returnStr
		if value != undefined then
			format "%" (RIBit_Functions.writeValueByType value brackets:true) to:returnStr
		returnStr as string
	)
) -- end of s_RIBit_Parameter struct


struct s_RIBit_Option (
	name,
	params = #(),
	indent = "",
	fileHandle,

	fn addParameter pname pvalue =
	(
		global s_RIBit_Parameter
		append params (s_RIBit_Parameter pname pvalue)
	),

	fn paramsFromText txt =
	(
		global RIBit_Functions
		local arr = filterString (txt as string) " "
		local tmpArr = #()
		local bracket = false
		local quote = false
		for i in arr do (
			if quote then (
				tmpArr[tmpArr.count] += " " + i
			) else if bracket then (
				tmpArr[tmpArr.count] += " " + i
			) else (
				if i[1] == "\"" then quote = true
				if i[1] == "[" then bracket = true
				append tmpArr i
			)
			if i[i.count] == "\"" then quote = false
			if i[i.count] == "]" then bracket = false
		)
		arr = tmpArr
		local i = 1
		while i <= arr.count do (
			if arr[i][1] == "\"" then (
				arr[i] = RIBit_Functions.replaceString arr[i] "\"" ""
				if i < arr.count and arr[i+1][1] != "\"" then (
					arr[i+1] = RIBit_Functions.replaceString arr[i+1] "[" ""
					arr[i+1] = RIBit_Functions.replaceString arr[i+1] "]" ""
					tmpArr = filterString arr[i+1] " "
					if tmpArr.count == 3 then (
						local allNumbers = true
						for i = 1 to 3 where allNumbers and (tmpArr[i] as float) == undefined do
							allNumbers = false
						if allNumbers then
							arr[i+1] = "[" + (RIBit_Functions.replaceString arr[i+1] " " ",") + "]"
					)
					try (arr[i+1] = execute arr[i+1])catch()
					addParameter arr[i] arr[i+1]
					i += 1
				) else (
					addParameter arr[i] undefined
				)
			)
			i += 1
		)
	),

	fn changeValue pname pvalue add:true =
	(
		global RIBit_Functions
		local names = for p in params collect (RIBit_Functions.toLowerCase p.name)
		local i = findItem names (RIBit_Functions.toLowerCase pname)
		if i > 0 then (
			params[i].value = pvalue
		) else if add == true then
			addParameter pname pvalue
	),

	fn write =
	(
		global RIBit_Functions
		if fileHandle != undefined then (
			format "%%" indent name to:fileHandle
			for p in params do
				format "%" (RIBit_Functions.writeValueByType p) to:fileHandle
			format "\n" to:fileHandle
		)
	)
) -- end of s_RIBit_Option srtuct



struct s_RIBit_Attribute (
	name,
	params = #(),
	indent = "",
	fileHandle,

	fn addParameter pname pvalue =
	(
		global s_RIBit_Parameter
		append params (s_RIBit_Parameter name:pname value:pvalue)
	),

	fn paramsFromText txt =
	(
		global RIBit_Functions
		local arr = filterString (txt as string) " "
		local tmpArr = #()
		local bracket = false
		local quote = false
		for i in arr do (
			if quote then (
				tmpArr[tmpArr.count] += " " + i
			) else if bracket then (
				tmpArr[tmpArr.count] += " " + i
			) else (
				if i[1] == "\"" then quote = true
				if i[1] == "[" then bracket = true
				append tmpArr i
			)
			if i[i.count] == "\"" then quote = false
			if i[i.count] == "]" then bracket = false
		)
		arr = tmpArr
		local i = 1
		while i <= arr.count do (
			if arr[i][1] == "\"" then (
				arr[i] = RIBit_Functions.replaceString arr[i] "\"" ""
				if i < arr.count and arr[i + 1][1] != "\"" then (
					arr[i + 1] = RIBit_Functions.replaceString arr[i + 1] "[" ""
					arr[i + 1] = RIBit_Functions.replaceString arr[i + 1] "]" ""
					try (arr[i + 1] = execute arr[i + 1])catch()
					addParameter arr[i] arr[i + 1]
					i += 1
				) else (
					addParameter arr[i] undefined
				)
			)
			i += 1
		)
	),

	fn changeValue pname pvalue add:true =
	(
		global RIBit_Functions
		local names = for p in params collect (RIBit_Functions.toLowerCase p.name)
		local i = findItem names (RIBit_Functions.toLowerCase pname)
		if i > 0 then (
			params[i].value = pvalue
		) else if add == true then
			addParameter pname pvalue
	),


	fn write =
	(
		global RIBit_Functions
		if fileHandle != undefined then (
			format "%Attribute \"%\"" indent name to:fileHandle
			for p in params do
				format "%" (RIBit_Functions.writeValueByType p) to:fileHandle
			format "\n" to:fileHandle
		)
	)
) -- end of s_RIBit_Attribute struct