--*********************************************************************************
--* This file is part of the RIBit Renderman(tm) exporter for max
--* By Ofer Zelichover
--* 2004/3/1 (c)
--* based in part on kcman by Rudy Cortes
--*********************************************************************************

struct s_RIBit_Particle (

	object,
	activeParticles = #{},
	camObject,
	settings = s_RIBit_Settings(),
	objInstances = #(),
	indent = "",
	fileHandle,
	
	
	
-- Functions
	fn incIndent =
	(
		indent += "\t"
	),

	fn decIndent =
	(
		if indent.count > 0 then
			indent = subString indent 1 (indent.count - 1)
	),

	fn getActiveParticles =
	(
		activeParticles = #{}
		local pCount = particleCount object
		for i = 1 to pCount do
			activeParticles[i] = particleAge object i != undefined
	),

	fn exportRIBBox =
	(
		global RIBit_Functions
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_RIBBox multiple:true
		if modIndex != undefined then (
			format "%##---- RIB Box ----##\n" indent to:fileHandle
			for i in modIndex do (
				local RIBFn = RIBit_Functions.parseRIBBoxAsFn object.modifiers[i].RIBCode
				local lines = filterString (RIBFn #(object)) "\n\r"
				for l in lines do
					format "%%\n" indent l to:fileHandle
			)
		)
	),


	fn getParticleColor i obj: =
	(
		global RIBit_Functions
		global g_RIBit_cacheFnColorPP
		if obj == unsupplied then obj = object
		local col = white
		local modIndex = RIBit_Functions.findModifier obj mo_RIBit_Particles
		if modIndex != undefined and obj.modifiers[modIndex].colorPP != "" then (
			if g_RIBit_cacheFnColorPP == undefined then
				g_RIBit_cacheFnColorPP = RIBit_Functions.parseRIBBoxAsFn obj.modifiers[modIndex].colorPP errorDefaultValue:white withErrorVar:false
			col = g_RIBit_cacheFnColorPP #(obj, i)
		) else if (modIndex = RIBit_Functions.findModifier obj mo_RIBit_Shaders) != undefined then (
			col = obj.modifiers[modIndex].Color
		)
		if col == undefined then 
			col = white
		if isKindOf col Integer or isKindOf col Float then
			col = color (255. * col) (255. * col) (255. * col)
		if not isKindOf col string then
			col = RIBit_Functions.writeValueByType col
		col
	),

	fn getParticleOpacity i obj: =
	(
		global RIBit_Functions
		global g_RIBit_cacheFnOpacityPP
		if obj == unsupplied then obj = object
		local op = white
		local modIndex = RIBit_Functions.findModifier obj mo_RIBit_Particles
		if modIndex != undefined and obj.modifiers[modIndex].opacityPP != "" then (
			if g_RIBit_cacheFnOpacityPP == undefined then
				g_RIBit_cacheFnOpacityPP = RIBit_Functions.parseRIBBoxAsFn obj.modifiers[modIndex].opacityPP errorDefaultValue:white withErrorVar:false
			op = g_RIBit_cacheFnOpacityPP #(obj, i)
		) else if (modIndex = RIBit_Functions.findModifier obj mo_RIBit_Shaders) != undefined then (
			op = obj.modifiers[modIndex].Opacity
		)
		if isKindOf op Integer or isKindOf op Float then
			op = color (255. * op) (255. * op) (255. * op)
		if not isKindOf op string then
			op = RIBit_Functions.writeValueByType op
		op
	),

	fn getParticleSize i obj: =
	(
		global RIBit_Functions
		global g_RIBit_cacheFnSizePP
		if obj == unsupplied then obj = object
		local s = 1.0
		local modIndex = RIBit_Functions.findModifier obj mo_RIBit_Particles
		if modIndex != undefined and obj.modifiers[modIndex].sizePP != "" then (
			if g_RIBit_cacheFnSizePP == undefined then
				g_RIBit_cacheFnSizePP = RIBit_Functions.parseRIBBoxAsFn obj.modifiers[modIndex].sizePP errorDefaultValue:white withErrorVar:false
			s = g_RIBit_cacheFnSizePP #(obj, i)
		) else if particleSize2 != undefined then
			s = particleSize2 obj i
		else
			s = particleSize obj i
		if not isKindOf s string then
			s = RIBit_Functions.writeValueByType s
		s
	),

	fn getParticleAge i obj: = 	-- get the particle normalized age, if available
	(
		if obj == unsupplied then obj = object
		if particleLife != undefined then
			(particleAge obj i) / (particleLife obj i)
		else
			(particleAge obj i)
	),

	fn exportPoints params: =
	(
		global RIBit_Functions
		local indent = params[1]
		local fileHandle = params[2]
		local object = params[3]
		local activeParticles = params[4]
		local getParticleSize = params[5]
		local getParticleColor = params[6]
		local getParticleOpacity = params[7]
		local getParticleAge = params[8]

		format "%Points\n" indent to:fileHandle

		indent += "\t"
		local ii = 1

		-- export positions
		format "%\"P\" [" indent to:fileHandle
		for i in activeParticles do (
			if ii > 8 then (
				format "\n%" indent to:fileHandle
				ii = 1
			)
			local p = in CoordSys object particlePos object i
			format " % % % " p.x p.y p.z to:fileHandle
			ii += 1
		)
		format "]\n" indent to:fileHandle

		-- export widths
		ii = 1
		format "%\"width\" [" indent to:fileHandle
		for i in activeParticles do (
			if ii > 24 then (
				format "\n%" indent to:fileHandle
				ii = 1
			)
			format " %" (getParticleSize i obj:object) to:fileHandle
			ii += 1
		)
		format "]\n" indent to:fileHandle

		-- export colors
		ii = 1
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Particles
		if modIndex != undefined and object.modifiers[modIndex].colorPP != "" then (
			format "%\"Cs\" [" indent to:fileHandle
			for i in activeParticles do (
				if ii > 8 then (
					format "\n%" indent to:fileHandle
					ii = 1
				)
				local Cs = (getParticleColor i obj:object)
				format " %" Cs to:fileHandle
--				format " % % %" (Cs.r / 255.) (Cs.g / 255.) (Cs.b / 255.) to:fileHandle
				ii += 1
			)
			format "]\n" indent to:fileHandle
		)

		-- export opacities
		ii = 1
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Particles
		if modIndex != undefined and object.modifiers[modIndex].opacityPP != "" then (
			format "%\"Os\" [" indent to:fileHandle
			for i in activeParticles do (
				if ii > 8 then (
					format "\n%" indent to:fileHandle
					ii = 1
				)
				local Os = (getParticleOpacity i obj:object)
				format " %" Os to:fileHandle
				ii += 1
			)
			format "]\n" indent to:fileHandle
		)

		-- export ages
		ii = 1
		format "%\"varying float age\" [" indent to:fileHandle
		for i in activeParticles do (
			if ii > 24 then (
				format "\n%" indent to:fileHandle
				ii = 1
			)
			local age = getParticleAge i obj:object
			format " %" (RIBit_Functions.frameToInt age) to:fileHandle
			ii += 1
		)
		format "]\n" indent to:fileHandle
	),


	fn exportMotionPoints =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		if (RIBit_Functions.findModifier object mo_RIBit_Shaders) == undefined then (
			local ribSurface = s_RIBit_Option "Surface" indent:indent fileHandle:fileHandle
			ribSurface.addParameter "constant" undefined
			ribSurface.write()
		)
		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		ribMotion.func = exportPoints
		ribMotion.funcParams = #(indent, fileHandle, object, activeParticles, getParticleSize, getParticleColor, getParticleOpacity, getParticleAge)
		ribMotion.dependsOn = #(true)
		ribMotion.write()
	),


	fn exportSpheres =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Particles

		local mat = matrix3 1
		local ribTransform = s_RIBit_Transform mat concat:true indent:indent fileHandle:fileHandle

		-- define the motion struct for transfrom
		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn motionFunc params: =
		(
			params[4].row4 = in CoordSys world particlePos params[2] params[3]
			params[1].transform = params[4] * (inverse params[2].transform)
			params[1].write()
		)

		ribMotion.func = motionFunc
		ribMotion.dependsOn = #(true)

		-- define the motion structs for color and opacity
		local ribColorMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn colorMotionFunc params: =
		(
			local obj = params[2]
			local particleID = params[3]
			local getParticleColOpFn = params[4]
			params[1].params = #()
			local col = (getParticleColOpFn particleID obj:obj)
			col = filterString col " "
			for i = 1 to col.count do
				col[i] = col[i] as float
			params[1].addParameter col undefined
			params[1].write()
		)
		ribColorMotion.func = colorMotionFunc
		ribColorMotion.dependsOn = #(true)
		ribColorMotion.bypass = settings.useConservativeMotion

		local ribColor = s_RIBit_Option "Color" indent:indent fileHandle:fileHandle

		local ribOpacity = s_RIBit_Option "Opacity" indent:indent fileHandle:fileHandle

		-- define the motion structs for the sphere object
		local ribSphereMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn sphereMotionFunc params: =
		(
			params[1].params = #()
			local r = try((params[4] params[3] obj:params[2]) as float)catch(particleSize params[2] params[3])
			params[1].addParameter r undefined
			params[1].addParameter -r undefined
			params[1].addParameter r undefined
			params[1].addParameter 360 undefined
			params[1].addParameter "uniform float age" (params[5] params[3] obj:params[2])
			params[1].write()
		)

		ribSphereMotion.func = sphereMotionFunc
		ribSphereMotion.dependsOn = #(true)

		local ribSphere = s_RIBit_Option "Sphere" indent:indent fileHandle:fileHandle

		-- loop through the particles and export the data
		for i in activeParticles do (
			ribTransform.transformBegin()
			indent = ribTransform.indent

			ribMotion.funcParams = #(ribTransform, object, i, mat)
			ribMotion.indent = indent
			ribMotion.write()


			if modIndex != undefined and object.modifiers[modIndex].colorPP != "" then (
				ribColorMotion.funcParams = #(ribColor, object, i, getParticleColor)
				ribColor.indent = ribColorMotion.indent = indent
				ribColorMotion.write()
			)
			if modIndex != undefined and object.modifiers[modIndex].opacityPP != "" then (
				ribColorMotion.funcParams = #(ribOpacity, object, i, getParticleOpacity)
				ribOpacity.indent = ribColorMotion.indent = indent
				ribColorMotion.write()
			)
			if modIndex != undefined and object.modifiers[modIndex].RIBBoxPP != "" then (
				local lines = filterString (RIBit_Functions.parseRIBBox object.modifiers[modIndex].RIBCode params:#(object, i)) "\n\r"
				for l in lines do
					format "%%\n" indent l to:fileHandle
			)

			ribSphereMotion.funcParams = #(ribSphere, object, i, getParticleSize, getParticleAge)
			ribSphere.indent = ribSphereMotion.indent = indent
			ribSphereMotion.write()

			ribTransform.transformEnd()
			indent = ribTransform.indent
			-- End Transformation block

		)
	),


	fn exportSprites =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local camDir = if isValidNode camObject then camObject.dir else (inverse(getViewTM())).row3
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Particles
		local ribTransform = s_RIBit_Transform concat:true indent:indent fileHandle:fileHandle

		-- calculate particle transformation so it has the same direction as the camera
		local Zvect = normalize (-camDir)
		local Xvect = normalize (cross [0,0,1] Zvect)
		local Yvect = normalize (cross Xvect Zvect)
		local mat = matrix3 -Xvect Yvect Zvect [0,0,0]

		-- define the motion struct for transfrom
		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn motionFunc params: =
		(
			params[4].row4 = in CoordSys world particlePos params[2] params[3]
			params[1].transform = params[4]
			if params[5] != undefined then (
				local twist = RIBit_Functions.parseRIBBox params[2].modifiers[params[5]].spriteTwistPP params:#(params[2], params[3])
				params[1].localRotateZ (try(twist as float)catch(0))
			)
			params[1].transform *= (inverse params[2].transform)
			params[1].write()
		)

		ribMotion.func = motionFunc
		ribMotion.dependsOn = #(true)

		-- define the motion structs for color and opacity
		local ribColorMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn colorMotionFunc params: =
		(
			local obj = params[2]
			local particleID = params[3]
			local getParticleColOpFn = params[4]
			params[1].params = #()
			local col = (getParticleColOpFn particleID obj:obj)
			col = filterString col " "
			for i = 1 to col.count do
				col[i] = col[i] as float
			params[1].addParameter col undefined
			params[1].write()
		)
		ribColorMotion.func = colorMotionFunc
		ribColorMotion.dependsOn = #(true)
		ribColorMotion.bypass = settings.useConservativeMotion

		local ribColor = s_RIBit_Option "Color" indent:indent fileHandle:fileHandle
		local ribOpacity = s_RIBit_Option "Opacity" indent:indent fileHandle:fileHandle


		-- define the motion structs for the sprite patch
		local ribSpriteMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn spriteMotionFunc params: =
		(
			params[1].params = #()
			params[1].addParameter "bilinear" undefined
			local r = try(execute (params[4] params[3] obj:params[2])) catch()
			local x, y
			if isKindOf r Point2 then (
				x = r.x
				y = r.y
			) else if (isKindOf r float) or (isKindOf r integer) then
				x = y = r / 2.
			else
				x = y = (particleSize params[2] params[3])
			params[1].addParameter "P" #(-x,-y,0, x,-y,0, -x,y,0, x,y,0)
			params[1].addParameter "uniform float age" (params[5] params[3] obj:params[2])
			params[1].write()
		)
		ribSpriteMotion.func = spriteMotionFunc
		ribSpriteMotion.dependsOn = #(true)

		local ribSprite = s_RIBit_Option "Patch" indent:indent fileHandle:fileHandle

		-- loop through the particles and export the data
		for i in activeParticles do (
			ribTransform.transformBegin()
			indent = ribTransform.indent


			ribMotion.funcParams = #(ribTransform, object, i, mat, modIndex)
			ribTransform.indent = ribMotion.indent = indent
			ribMotion.write()

			if modIndex != undefined and object.modifiers[modIndex].colorPP != "" then (
				ribColorMotion.funcParams = #(ribColor, object, i, getParticleColor)
				ribColor.indent = ribColorMotion.indent = indent
				ribColorMotion.write()
			)
			if modIndex != undefined and object.modifiers[modIndex].opacityPP != "" then (
				ribColorMotion.funcParams = #(ribOpacity, object, i, getParticleOpacity)
				ribOpacity.indent = ribColorMotion.indent = indent
				ribColorMotion.write()
			)

			if modIndex != undefined and object.modifiers[modIndex].RIBBoxPP != "" then (
				local lines = filterString (RIBit_Functions.parseRIBBox object.modifiers[modIndex].RIBCode params:#(object, i)) "\n\r"
				for l in lines do
					format "%%\n" indent l to:fileHandle
			)

			ribSpriteMotion.funcParams = #(ribSprite, object, i, getParticleSize, getParticleAge)
			ribSprite.indent = ribSpriteMotion.indent = indent
			ribSpriteMotion.write()

			ribTransform.transformEnd()
			indent = ribTransform.indent
			-- End Transformation block

		)
	),


	fn exportInstancedObjects objs =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local ribObject = s_RIBit_Object indent:indent fileHandle:fileHandle
		for o in objs where ((findItem objInstances o == 0) and isKindOf o GeometryClass and not (isKindOf o targetObject)) do (
			append objInstances o
			ribObject.id = objInstances.count
			ribObject.ObjectBegin()
			indent = ribObject.indent

			local ribGeometry = s_RIBit_Geometry object:o settings:settings indent:indent fileHandle:fileHandle
			ribGeometry.exportGeometry()

			ribObject.ObjectEnd()
			indent = ribObject.indent
		)
	),

	fn exportInstances =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Particles
		local instObjs = object.modifiers[modIndex].instanceObjects
		if instObjs.count == 0 then (
			exportMotionPoints()
			return()
		)

		exportInstancedObjects instObjs

		instObjs = for o in instObjs collect #(o,(findItem objInstances o))
		local lastUsedInst = 1


		local orientMode = 1
		local mat
		if modIndex != undefined then
			orientMode = object.modifiers[modIndex].instanceOrientMode

		-- define the motion struct for transfrom
		local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn motionFunc params: =
		(
			global RIBit_Functions
			local ribTrasform = params[1]; local object = params[2]; local i = params[3]; local orientMode = params[4]
			local modIndex = params[5]; local getParticleSize = params[6]; local camObject = params[7]; local instObj = params[8]
			-- calculate particle transformation so it has the same direction as the camera
			local camDir = if isValidNode camObject then camObject.dir else (inverse(getViewTM())).row3
			local Zvect = normalize (-camDir)
			local Xvect = normalize (cross [0,0,1] Zvect)
			local Yvect = normalize (cross Xvect Zvect)
			local mat
			case orientMode of (
				1: mat = (in CoordSys world instObj.transform)
				2: mat = (random (eulerAngles 0 0 0) (eulerAngles 360. 360. 360.)) as matrix3
				3: mat = matrix3 -Xvect Yvect Zvect [0,0,0]
				4: if object.modifiers[modIndex].instanceOrientPP != "" then(
						rot = RIBit_Functions.parseRIBBox object.modifiers[modIndex].instanceOrientPP params:#(object, i)
						rot = try(execute rot)catch()
						if isKindOf rot Point3 then rot = eulerAngles rot[1] rot[2] rot[3]
						mat = try(rot as matrix3)catch(matrix3 1)
					) else mat = matrix3 1
				default: mat = matrix3 1
			)
			local s = try(execute (getParticleSize i obj:object))catch()
			if isKindOf s Float or isKindOf s Integer then s = [s,s,s]
			if not isKindOf s Point3 then s = [1,1,1]
			ribTrasform.transform = mat
			ribTrasform.scale s
			ribTrasform.transform.row4 = in CoordSys world particlePos object i

			if orientMode == 3 then (
				local twist = RIBit_Functions.parseRIBBox object.modifiers[modIndex].instanceOrientPP params:#(object, i)
				ribTrasform.localRotateZ (try(twist as float)catch(0))
			)
			ribTrasform.transform *= (inverse object.transform)
			ribTrasform.write()
		)

		ribMotion.func = motionFunc
		ribMotion.dependsOn = #(true)

		local ribTransform = s_RIBit_Transform concat:true indent:indent fileHandle:fileHandle

		-- define the motion structs for color and opacity
		local ribColorMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
		fn colorMotionFunc params: =
		(
			global RIBit_Functions
			local obj = params[2]
			local particleID = params[3]
			local getParticleColOpFn = params[4]
			params[1].params = #()
			if params[5] != undefined then
				params[1].addParameter params[5] undefined
			local col = (getParticleColOpFn particleID obj:obj)
			if isKindOf col string then (
				col = filterString col " "
				for i = 1 to col.count do
					col[i] = col[i] as float
			)
			params[1].addParameter col undefined
			params[1].write()
		)
		ribColorMotion.func = colorMotionFunc
		ribColorMotion.dependsOn = #(true)
		ribColorMotion.bypass = settings.useConservativeMotion

		local ribColor = s_RIBit_Option "Color" indent:indent fileHandle:fileHandle
		local ribOpacity = s_RIBit_Option "Opacity" indent:indent fileHandle:fileHandle

		-- define the motion structs for the age attribute
		local ribAge = s_RIBit_Attribute "user" indent:indent fileHandle:fileHandle

		-- define the motion structs for the instanced object
		local ribInstance = s_RIBit_Option "ObjectInstance" indent:indent fileHandle:fileHandle

	-- loop through the particles and export the data
		for i in activeParticles do (
			-- get the instanced object id
			local instID = lastUsedInst
			if modIndex != undefined and object.modifiers[modIndex].instanceMode == 2 then (
				lastUsedInst += 1
				if lastUsedInst > instObjs.count then
					lastUsedInst = 1
			) else (
				seed i
				instID = random 1 instObjs.count
			)

			ribTransform.transformBegin()
			indent = ribTransform.indent
 			ribMotion.funcParams = #(ribTransform, object, i, orientMode, modIndex, getParticleSize, camObject, instObjs[instID][1])
			ribTransform.indent = ribMotion.indent = indent
			ribMotion.write()

			if modIndex != undefined and object.modifiers[modIndex].colorPP != "" then (
				ribColorMotion.funcParams = #(ribColor, object, i, getParticleColor)
				ribColor.indent = ribColorMotion.indent = indent
				ribColorMotion.write()
			)
			if modIndex != undefined and object.modifiers[modIndex].opacityPP != "" then (
				ribColorMotion.funcParams = #(ribOpacity, object, i, getParticleOpacity)
				ribOpacity.indent = ribColorMotion.indent = indent
				ribColorMotion.write()
			)

			-- export the age attribute
			ribColorMotion.funcParams = #(ribAge, object, i, getParticleAge, "uniform float age")
			ribAge.indent = ribColorMotion.indent = indent
			ribColorMotion.write()

			if modIndex != undefined and object.modifiers[modIndex].RIBBoxPP != "" then (
				local lines = filterString (RIBit_Functions.parseRIBBox object.modifiers[modIndex].RIBCode params:#(object, i)) "\n\r"
				for l in lines do
					format "%%\n" indent l to:fileHandle
			)

			ribInstance.params = #()
			ribInstance.addParameter instObjs[instID][2] undefined
			ribInstance.indent = indent
			ribInstance.write()

			ribTransform.transformEnd()
			indent = ribTransform.indent
			-- End Transformation block
		)
	),


	fn write =
	(
		global RIBit_Functions
		global g_RIBit_cacheFnColorPP = undefined
		global g_RIBit_cacheFnOpacityPP = undefined
		global g_RIBit_cacheFnSizePP = undefined
		if fileHandle == undefined then return()

		getActiveParticles()

		local modIndex = RIBit_Functions.findModifier object mo_RIBit_Particles
		if modIndex != undefined then (
			case object.modifiers[modIndex].type of (
				1: exportMotionPoints()
--				2: exportCurves()
				3: exportSpheres()
--				4: exportBlobbies()
				5: exportSprites()
				6: exportInstances()
				default: exportMotionPoints()
			)
		) else
			exportMotionPoints()
		
		g_RIBit_cacheFnColorPP = undefined
		g_RIBit_cacheFnOpacityPP = undefined
		g_RIBit_cacheFnSizePP = undefined
	)
) -- end of s_RIBit_Particle struct


/*

particleSize (const NodeGeneric): particleSize()
particleAge (const NodeGeneric): particleAge()
particleVelocity (const NodeGeneric): particleVelocity()
particlePos (const NodeGeneric): particlePos()
particleCount (const NodeGeneric): particleCount()

*/