--*********************************************************************************
--* This file is part of the RIBit Renderman(tm) exporter for max
--* By Ofer Zelichover
--* 2004/3/1 (c)
--* based in part on kcman by Rudy Cortes

--*********************************************************************************

struct s_RIBit_Scene (
	type = #scene,
	RIBFilename,
	renderCamera,
	settings = s_RIBit_Settings(),
	shadersToCompile = #(),
	shadowmaps = #(),
	objInstances = #(),
	indent = "",
	fileHandle,
	canceled = false,
	progressCount = 0,
	exportTime = 0,

	fn incIndent =
	(
		indent += "\t"
	),

	fn decIndent =
	(
		if indent.count > 0 then
			indent = subString indent 1 (indent.count - 1)
	),

	fn formatPath path =
	(
		if settings.activeRenderer.useForwardSlash then
			RIBit_Functions.winToUnixPath path
		else
			RIBit_Functions.replaceString (RIBit_Functions.unixToWinPath path) "\\" "\\\\"
	),

	fn writeComment comment indent:indent =
	(
		if fileHandle == undefined then return()
		comment = comment as string
		format "%#%\n" indent comment to:fileHandle
	),

	fn writeHeader =
	(
		writeComment "# RIB file structure"
		writeComment ("# Generated using RIBit ver. " + settings.version as string + " by Ofer Zelichover")
		writeComment "# Base on kcrib by Rudi Cortes"
		writeComment ("# Created on " + localtime + " by " + sysInfo.username + " on 3DSMax ver. " \
						+ (maxversion() as string) + "\n")
	),

	fn writeSearchPaths =
	(
		format "%Option \"searchpath\" \"shader\" \"%:%:&\"\n" indent (formatPath (settings.BaseRIBitDir + "Shaders\\")) (formatPath settings.ShaderDir) to:fileHandle
		format "%Option \"searchpath\" \"texture\" \"%:&\"\n" indent (formatPath settings.TextureDir) to:fileHandle
		format "\n" to:fileHandle
	),


	fn writeUserOptions =
	(
		local usedUserOps = false
		if type == #shadowmap then (
			format "%Option \"user\" \"string pass\" \"shadowmap\"\n" indent to:fileHandle
			usedUserOps = true
		)
		if usedUserOps then
			format "\n" indent to:fileHandle
	),


	fn writeDisplay =
	(
		global s_RIBit_Display
		local n = settings.dispName
		if type == #shadowmap then
			n += ".depth"
		local ribDisplay= s_RIBit_Display name:n type:settings.dispType channels:settings.dispChannels fileHandle:fileHandle indent:indent
		ribDisplay.write()
	),

	fn writeFormat =
	(
		global s_RIBit_Option
		local ribFormat = s_RIBit_Option "Format" fileHandle:fileHandle indent:indent
		ribFormat.addParameter settings.rendWidth undefined
		ribFormat.addParameter settings.rendHeight undefined
		ribFormat.addParameter settings.rendPixelAspect undefined
		ribFormat.write()
	),

	fn writeShutter =
	(
		global s_RIBit_Option
		if settings.useMotionBlur then (
			local ribShutter = s_RIBit_Option "Shutter" fileHandle:fileHandle indent:indent
			ribShutter.addParameter 0.0 undefined
			ribShutter.addParameter (settings.shutterAngle / 360.) undefined
			ribShutter.write()
		)
	),

	fn writeFilter =
	(
		global s_RIBit_Option
		local ribFilter = s_RIBit_Option "PixelFilter" fileHandle:fileHandle indent:indent
		ribFilter.addParameter settings.filter[1] undefined
		ribFilter.addParameter settings.filter[2] undefined
		ribFilter.addParameter settings.filter[3] undefined
		ribFilter.write()

		local ribPixelSamples = s_RIBit_Option "PixelSamples" fileHandle:fileHandle indent:indent
		ribPixelSamples.addParameter settings.pixelSamples[1] undefined
		ribPixelSamples.addParameter settings.pixelSamples[2] undefined
		ribPixelSamples.write()

		local ribShadingRate = s_RIBit_Option "ShadingRate" fileHandle:fileHandle indent:indent
		ribShadingRate.addParameter settings.shadingRate undefined
		ribShadingRate.write()
	),


	fn writeBackground =
	(
		global s_RIBit_Option
		if backgroundColor != black then (
			local ribImager = s_RIBit_Option "Imager" fileHandle:fileHandle indent:indent
			ribImager.addParameter "background" undefined
			ribImager.addParameter "color bgcolor" #((backgroundColor.r/255), (backgroundColor.g/255), (backgroundColor.b/255))
			ribImager.write()
		)
	),


	fn writeRenderOptions =
	(
		global s_RIBit_Option
		if findItem settings.RenderOptions #force2sided > 0 then (
			local ribOption = s_RIBit_Option "Sides" fileHandle:fileHandle indent:indent
			ribOption.addParameter 2 undefined
			ribOption.write()
		)
	),


	fn exportLights =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		writeComment "# Light Declarations"
		-- export ambient light, if present.
		if ambientColor != black then (
			local ribMotion = s_RIBit_Motion settings:settings indent:indent fileHandle:fileHandle
			local ribLight = s_RIBit_Light ambientColor lightID:0 shadersDir:settings.ShaderDir indent:indent fileHandle:fileHandle

			fn motionFunc params: =
			(
				params[1].light = ambientColor
				params[1].write()
			)
			ribMotion.func = motionFunc
			ribMotion.funcParams = #(ribLight)
			ribMotion.dependsOn = #(ambientColorController.keys.count > 1)
			ribMotion.bypass = settings.useConservativeMotion
			ribMotion.write()
		)

		local allLights = for l in lights where (not isKindOf l Targetobject) collect l
		if allLights.count == 0 then
			writeComment "# No Lights on this scene\n"
		else (
			local shdmapLights = for o in shadowmaps collect o.object
			progressStart "RIBit: Exporting lights..."
			for i = 1 to allLights.count where (not canceled) do (
				canceled = not (progressUpdate (100 * ((i-1.)/allLights.count as float)))
				local ribLight = s_RIBit_Light allLights[i] lightID:i shadersDir:settings.ShaderDir  \
										 shadowSamples:(32 / settings.shadingRate) settings:settings indent:indent fileHandle:fileHandle
				local shdmapIndex = findItem shdmapLights allLights[i]
				if shdmapIndex > 0 then
					ribLight.shadowmapName = shadowmaps[shdmapIndex].name + ".tx"
				ribLight.write()

				RIBit_Functions.uniqueJoin shadersToCompile ribLight.shadersToComplie
			)
			progressEnd()
		)
	),


	fn exportObjects obj totalcnt:0 =
	(
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		canceled = getProgressCancel()
		if obj == undefined or canceled then return()
		local needExport = isKindOf obj GeometryClass and not (isKindOf obj targetObject) and obj.renderable and not obj.isHidden
		needExport = needExport or (obj.children.count > 0)
		needExport = needExport and obj != rootNode
		if needExport and not canceled then (
		local progVal = (100 * (progressCount / totalcnt)) as integer
			progVal = amax progVal 1
			canceled = not (progressUpdate progVal)
			progressCount += 1

			local ribObject = s_RIBit_Geometry obj shadersDir:settings.ShaderDir camObject:renderCamera objInstances:objInstances \
													canceled:canceled settings:settings indent:(indent + "\t") fileHandle:fileHandle
			ribObject.force2sided = (findItem settings.renderOptions #force2sided) > 0
			writeComment ("#---- " + obj.name + " decalration starts ----##")
			format "%AttributeBegin\n" indent to:fileHandle

			ribObject.write()

			RIBit_Functions.uniqueJoin shadersToCompile ribObject.shadersToComplie
		)
		incIndent()
		for o in obj.children where (not canceled) do (
			exportObjects o totalcnt:totalcnt
		)
		decIndent()
		if needExport then (
			format "%AttributeEnd\n" indent to:fileHandle
			writeComment ("#---- " + obj.name + " decalration ends ----##")
		)

	),

	fn genRIBFileName =
	(
		global RIBit_Functions
		local RIBType = ""
		case type of (
			#shadowmap: RIBType = "SHD_" + renderCamera.name + "_"
		)

		local fname = getFilenameFile MaxFilename
		if fname == "" then
			fname = "Untitled"
		local frameStr = RIBit_Functions.leadingZero (RIBit_Functions.frameToInt settings.frame) 4

		RIBFilename = settings.RIBDir + RIBType + fname + "_" + frameStr + ".rib"
	),

	fn exportScene =
	(
		local t = timeStamp()
		global s_RIBit_Functions, s_RIBit_Parameter, s_RIBit_Option, s_RIBit_Attribute, s_RIBit_Motion, s_RIBit_ShaderInfoParam, s_RIBit_ShaderInfo
		global s_RIBit_ShaderInfoParser, s_RIBit_Transform, s_RIBit_Settings, s_RIBit_Display, s_RIBit_Camera, s_RIBit_Frame, s_RIBit_World, s_RIBit_Object
		global s_RIBit_Light, s_RIBit_Shadowmap, s_RIBit_Particle, s_RIBit_Geometry, s_RIBit_Scene
		global RIBit_Functions
		if RIBFilename == undefined or RIBFilename == "" or type != #scene then
			genRIBFileName()
		fileHandle = createFile RIBFilename

		if fileHandle == undefined then return()

		-- Write some general info
		writeHeader()
		writeSearchPaths()
		writeUserOptions()


		local ribFrame = s_RIBit_Frame frame:settings.frame fileHandle:fileHandle
		ribFrame.FrameBegin()
		incIndent()

		-- Export Display information
		writeComment "# Image Display Declaration" to:fileHandle
		writeDisplay()
		writeFormat()
		writeShutter()
		writeFilter()
		if type == #scene then (
			writeBackground()
		)
		writeRenderOptions()

		-- Export Camera Information
		local ribCamera = s_RIBit_Camera cam:renderCamera shadersDir:settings.ShaderDir settings:settings fileHandle:fileHandle indent:indent
		ribCamera.write()

		RIBit_Functions.uniqueJoin shadersToCompile ribCamera.shadersToComplie

		-- World Begin
		local ribWorld = s_RIBit_World indent:indent fileHandle:fileHandle
		ribWorld.WorldBegin()

		incIndent()
		if type == #scene then
			exportLights()
		decIndent()

		progressStart "RIBit: Exporting objects..."
		progressCount = 0
		exportObjects rootNode totalcnt:(geometry.count as float)
		progressEnd()

		ribWorld.WorldEnd()
		decIndent()

		if type == #shadowmap then (
			local shadowName = formatPath (settings.textureDir + settings.dispName + ".tx")
			format "%MakeShadow \"%\" \"%\"\n" indent (settings.dispName + ".depth") shadowName to:fileHandle
		)

		ribFrame.FrameEnd()

		close fileHandle
		resumeEditing()
		
		exportTime = timeStamp() - t
		RIBFilename
	)

) -- end of s_RIBit_Scene

