--*********************************************************************************
--* This file is part of the RIBit Renderman(tm) exporter for max
--* By Ofer Zelichover
--* 2004/3/1 (c)
--* based in part on kcman by Rudy Cortes --*********************************************************************************

struct s_RIBit_ShaderInfoParam (
	name,
	class,
	type,
	defaultValue,
	value,
	
	fn isAllNumbersArray arr =
	(
		local allNumbers = true
		for i = 1 to arr.count where allNumbers and (try((arr[i] as float) == undefined)catch(true)) do
			allNumbers = false
		allNumbers
	),
	
	fn validateValue =
	(
		if value != undefined then (
			case type of (
				"string": 	if not isKindOf value string then value = value as string else value = undefined
				"float": 	if not isKindOf value float then try(value = value as float)catch(value = undefined) else value = undefined
				"color": 	if not isKindOf value color then (
					local tmpArr = filterString (value as string) ", ;:"
					local allNumbers = isAllNumbersArray tmpArr
					if allNumbers and tmpArr.count == 1 then
						value = tmpArr as float
					else if allNumbers and tmpArr.count == 3 then
						value = color (tmpArr[1] as float) (tmpArr[2] as float) (tmpArr[3] as float)
					else
						value = undefined
				)
				"vector": 	if not isKindOf value point3 then (
					local tmpArr = filterString (value as string) ", ;:"
					local allNumbers = isAllNumbersArray tmpArr
					if allNumbers and tmpArr.count == 3 then
						value = point3 (tmpArr[1] as float) (tmpArr[2] as float) (tmpArr[3] as float)
					else
						value = undefined
				)
				"point": 	if not isKindOf value point3 then (
					local tmpArr = filterString (value as string) ", ;:"
					local allNumbers = isAllNumbersArray tmpArr
					if allNumbers and tmpArr.count == 3 then
						value = point3 (tmpArr[1] as float) (tmpArr[2] as float) (tmpArr[3] as float)
					else
						value = undefined
				)
				"normal": 	if not isKindOf value point3 then (
					local tmpArr = filterString (value as string) ", ;:"
					local allNumbers = isAllNumbersArray tmpArr
					if allNumbers and tmpArr.count == 3 then
						value = point3 (tmpArr[1] as float) (tmpArr[2] as float) (tmpArr[3] as float)
					else
						value = undefined
				)
			)
		)
	)
) -- end of s_RIBit_ShaderInfoParam struct


struct s_RIBit_ShaderInfo (
	name,
	type,
	params = #(),

	fn asString =
	(
		global RIBit_Functions
		local ss = "" as stringStream
		for p in params do (
			p.validateValue()
			if p.value != undefined and p.value != p.defaultValue and p.name != undefined then (
				format "\"" to:ss
				if p.class != undefined then
					format "% " p.class to:ss
				if p.type != undefined then
					format "% " p.type to:ss
				format "%\"" p.name to:ss
				format "% " (RIBit_Functions.writeValueByType p.value brackets:true) to:ss
			)
		)
		ss as string
	)
) -- end of s_RIBit_ShaderInfo struct




struct s_RIBit_ShaderInfoParser (
	getInfoCommand,
	shaderExt,
	tempFile = sysInfo.tempDir + "RIBit_shader_info.tmp",
	parseInfoFn,

	fn getInfo shaderName =
	(
		global RIBit_Functions
		local curDir = sysInfo.currentdir
		sysInfo.currentdir = getFilenamePath shaderName
		shaderName = getFilenameFile shaderName
		dosCommand (getInfoCommand + " " + shaderName + " >" + tempFile)
		local f = openFile tempFile mode:"r"
		local shaderInfo = s_RIBit_ShaderInfo()

		if f != undefined then (
			try (
				if parseInfoFn != undefined then
					parseInfoFn &f &shaderInfo
			) catch()
			close f
		)
		deleteFile tempFile
		sysInfo.currentdir = curDir
		shaderInfo
	)
) -- end of s_RIBit_ShaderInfoParser struct



